/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Admin - DataCiteImportService.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.service.metadata.imports;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;
import ch.unige.solidify.util.XMLTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.ContributorService;
import ch.unige.aou.business.LicenseService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.tool.CleanTool;
import ch.unige.aou.model.xml.deposit.v2_4.Contributor;
import ch.unige.aou.model.xml.deposit.v2_4.DateTypes;
import ch.unige.aou.model.xml.deposit.v2_4.DepositDoc;
import ch.unige.aou.model.xml.deposit.v2_4.Text;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.rest.JournalTitleRemoteService;

@Service
@ConditionalOnBean(AdminController.class)
public class DataCiteImportService extends MetadataImportService {

  private static final String DATACITE_SCHEMA_2_2 = "http://datacite.org/schema/kernel-2.2";
  private static final String DATACITE_SCHEMA_3 = "http://datacite.org/schema/kernel-3";
  private static final String DATACITE_SCHEMA_4 = "http://datacite.org/schema/kernel-4";
  private static final String XPATH_DOI = "/d:resource/d:identifier[@identifierType=\"DOI\"]";
  private static final String XPATH_TITLE = "/d:resource/d:titles/d:title";
  private static final String XPATH_CREATORS = "/d:resource/d:creators/d:creator";
  private static final String XPATH_ABSTRACT = "/d:resource/d:descriptions/d:description[@descriptionType=\"Abstract\"]";
  private static final String XPATH_PUBLICATION_YEAR = "/d:resource/d:publicationYear";
  private static final String XPATH_LANGUAGE = "/d:resource/d:language";
  private static final String XPATH_SUBJECTS = "/d:resource/d:subjects/d:subject";
  private static final String XPATH_TYPES = "/d:resource/d:resourceType/@resourceTypeGeneral";

  private final String apiUrl;

  public DataCiteImportService(AouProperties aouProperties,
          MessageService messageService, MetadataService metadataService, ContributorService contributorService,
          JournalTitleRemoteService journalTitleRemoteService, LicenseService licenseService) {
    super(aouProperties, messageService, metadataService, contributorService, journalTitleRemoteService, licenseService);
    this.apiUrl = aouProperties.getMetadata().getImports().getDataCite().getUrl();
  }

  public DepositDoc searchDoi(String doi, DepositDoc depositDoc) {
    try {
      final JsonNode parentNode = this.getJsonFromDataCite(doi);
      final Document dataCiteDoc = this.getXmlFromJson(parentNode);
      final String schemaVersion = this.getSchemaVersionFromJson(parentNode);
      this.fillDepositDoc(doi, depositDoc, dataCiteDoc, schemaVersion);
      return depositDoc;
    } catch (final XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by Datacite", e);
    }
  }

  public void fillDepositDoc(String doi, DepositDoc depositDoc, Document dataciteDoc, String schemaVersion) throws XPathExpressionException {
    final Map<String, String> namespace;
    switch (schemaVersion) {
      case "4":
        namespace = Map.of("d", DATACITE_SCHEMA_4);
        break;
      case "3":
        namespace = Map.of("d", DATACITE_SCHEMA_3);
        break;
      case "2.2":
        namespace = Map.of("d", DATACITE_SCHEMA_2_2);
        break;
      default:
        throw new SolidifyRuntimeException("Datacite version: " + schemaVersion + " not supported");
    }
    final XPath xpath = XMLTool.buildXPath(namespace);

    Node doiNode = (Node) xpath.compile(XPATH_DOI).evaluate(dataciteDoc, XPathConstants.NODE);
    Node titleNode = (Node) xpath.compile(XPATH_TITLE).evaluate(dataciteDoc, XPathConstants.NODE);
    Node creatorNode = (Node) xpath.compile(XPATH_CREATORS).evaluate(dataciteDoc, XPathConstants.NODE);
    Node abstractNode = (Node) xpath.compile(XPATH_ABSTRACT).evaluate(dataciteDoc, XPathConstants.NODE);
    Node publicationYearNode = (Node) xpath.compile(XPATH_PUBLICATION_YEAR).evaluate(dataciteDoc, XPathConstants.NODE);
    Node languageNode = (Node) xpath.compile(XPATH_LANGUAGE).evaluate(dataciteDoc, XPathConstants.NODE);
    Node subjectNode = (Node) xpath.compile(XPATH_SUBJECTS).evaluate(dataciteDoc, XPathConstants.NODE);
    Node resourceTypeNode = (Node) xpath.compile(XPATH_TYPES).evaluate(dataciteDoc, XPathConstants.NODE);

    if (doiNode != null) {
      this.fillDoi(depositDoc, dataciteDoc, XPATH_DOI, namespace);
    }
    if (titleNode != null) {
      this.fillTitle(depositDoc, dataciteDoc, XPATH_TITLE, namespace);
    }
    if (creatorNode != null) {
      this.fillContributors(depositDoc, dataciteDoc, namespace);
    }
    if (abstractNode != null) {
      this.fillAbstracts(depositDoc, dataciteDoc, namespace);
    }
    if (publicationYearNode != null) {
      this.fillPublicationYear(depositDoc, dataciteDoc, namespace);
    }
    if (languageNode != null) {
      this.fillLanguage(depositDoc, dataciteDoc, namespace);
    }
    if (subjectNode != null) {
      this.fillKeywords(depositDoc, dataciteDoc, namespace);
    }
    if (resourceTypeNode != null) {
      this.fillPublicationTypeAndSubtype(depositDoc, dataciteDoc, namespace);
    }

    // if the DOI fetched from metadata is different from the one used to retrieve metadata, replace it
    this.ensureDepositIdentifiersExist(depositDoc);
    if (StringTool.isNullOrEmpty(depositDoc.getIdentifiers().getDoi()) || !depositDoc.getIdentifiers().getDoi().equals(doi)) {
      depositDoc.getIdentifiers().setDoi(doi);
    }
  }

  private JsonNode getJsonFromDataCite(String doi) {
    try {
      RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
      RestTemplate client = restTemplateBuilder.rootUri(this.apiUrl).build();
      Map<String, String> parameters = new HashMap<>();
      parameters.put("doi", doi);
      final String jsonString = client.getForObject(this.apiUrl, String.class, parameters);
      return new ObjectMapper().readTree(jsonString);
    } catch (HttpClientErrorException e) {
      throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
              this.messageService.get("deposit.error.identifiers.doi.not_found_by_datacite", new Object[] { doi }));
    } catch (IOException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by DataCite", e);
    }
  }

  private Document getXmlFromJson(JsonNode parentNode) {
    final String encodedXmlString = parentNode.get("data").get("attributes").get("xml").textValue();
    final String xmlString = new String(Base64.getDecoder().decode(encodedXmlString), StandardCharsets.UTF_8);
    try {
      return XMLTool.parseXML(xmlString);
    } catch (ParserConfigurationException | SAXException | IOException e) {
      throw new SolidifyRuntimeException("Unable to search DOI by DataCite", e);
    }
  }

  private String getSchemaVersionFromJson(JsonNode parentNode) {
    return parentNode.get("data").get("attributes").get("schema-version").textValue();
  }

  private void fillPublicationTypeAndSubtype(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {

    String value = XMLTool.getFirstNodeInnerRawContent(dataciteDoc, XPATH_TYPES, namespaces);

    if (!StringTool.isNullOrEmpty(value)) {
      String subType = this.getPublicationSubType(value);
      String type = this.getPublicationType(value);
      depositDoc.setSubtype(subType);
      depositDoc.setType(type);
    }
  }

  private void fillAbstracts(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {
    String value = XMLTool.getFirstNodeInnerRawContent(dataciteDoc, XPATH_ABSTRACT, namespaces);

    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureAbstractsExist(depositDoc);
      Text text = this.getText(value);
      depositDoc.getAbstracts().getAbstract().add(text);
    }
  }

  private void fillContributors(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {
    final XPath xpath = XMLTool.buildXPath(namespaces);
    try {
      NodeList authorsList = (NodeList) xpath.compile(XPATH_CREATORS).evaluate(dataciteDoc, XPathConstants.NODESET);

      if (authorsList.getLength() > 0) {
        this.ensureContributorsExist(depositDoc);
      }

      for (int i = 0; i < this.getNumberMaxContributor(authorsList.getLength()); i++) {
        String familyName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:familyName", namespaces);
        String givenName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:givenName", namespaces);

        if (familyName == null && givenName == null) {
          String creatorName = XMLTool.getFirstTextContent(authorsList.item(i), "./d:creatorName", namespaces);
          if (creatorName != null && creatorName.contains(",")) {
            familyName = creatorName.split(",")[0].trim();
            givenName = creatorName.split(",")[1].trim();
          }
        }

        familyName = this.getNameOrUndefinedValue(familyName);
        givenName = this.getNameOrUndefinedValue(givenName);

        Contributor contributor = new Contributor();
        contributor.setLastname(CleanTool.capitalizeString(familyName));
        contributor.setFirstname(CleanTool.capitalizeString(givenName));
        depositDoc.getContributors().getContributorOrCollaboration().add(contributor);
      }
    } catch (XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to fill contributors", e);
    }
  }

  private void fillPublicationYear(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {
    String value = XMLTool.getFirstNodeInnerRawContent(dataciteDoc, XPATH_PUBLICATION_YEAR, namespaces);

    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureDatesExist(depositDoc);
      depositDoc.getDates().getDate().add(this.getDateWithType(DateTypes.PUBLICATION, value));
    }
  }

  private void fillLanguage(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {
    String value = XMLTool.getFirstTextContent(dataciteDoc, XPATH_LANGUAGE, namespaces);
    if (!StringTool.isNullOrEmpty(value)) {
      this.ensureLanguagesExist(depositDoc);
      depositDoc.getLanguages().getLanguage().add(this.getLanguageCode(value));
      if (depositDoc.getTitle() != null) {
        depositDoc.getTitle().setLang(this.getLanguageCode(value));
      }
    }
  }

  private void fillKeywords(DepositDoc depositDoc, Document dataciteDoc, Map<String, String> namespaces) {
    final XPath xpath = XMLTool.buildXPath(namespaces);
    try {
      NodeList keywordsList = (NodeList) xpath.compile(XPATH_SUBJECTS).evaluate(dataciteDoc, XPathConstants.NODESET);

      if (keywordsList.getLength() > 0) {
        this.ensureKeywordsExist(depositDoc);
      }

      for (int i = 0; i < keywordsList.getLength(); i++) {
        String keyword = keywordsList.item(i).getTextContent();
        depositDoc.getKeywords().getKeyword().add(keyword);
      }

    } catch (XPathExpressionException e) {
      throw new SolidifyRuntimeException("Unable to fill keywords", e);
    }
  }

  private String getPublicationType(String value) {
    switch (value) {
      case "Book":
      case "BookChapter":
        return AouConstants.DEPOSIT_TYPE_LIVRE_NAME;
      case "ConferencePaper":
      case "ConferenceProceeding":
        return AouConstants.DEPOSIT_TYPE_CONFERENCE_NAME;
      case "DataPaper":
      case "JournalArticle":
        return AouConstants.DEPOSIT_TYPE_ARTICLE_NAME;
      case "Dissertation":
        return AouConstants.DEPOSIT_TYPE_DIPLOME_NAME;
      case "Journal":
        return AouConstants.DEPOSIT_TYPE_JOURNAL_NAME;
      case "Preprint":
      case "Report":
        return AouConstants.DEPOSIT_TYPE_RAPPORT_NAME;
      default:
        return null;
    }
  }

  private String getPublicationSubType(String value) {
    switch (value) {
      case "Book":
        return AouConstants.DEPOSIT_SUBTYPE_LIVRE_NAME;
      case "BookChapter":
        return AouConstants.DEPOSIT_SUBTYPE_CHAPITRE_DE_LIVRE_NAME;
      case "ConferencePaper":
        return AouConstants.DEPOSIT_SUBTYPE_ACTES_DE_CONFERENCE_NAME;
      case "ConferenceProceeding":
        return AouConstants.DEPOSIT_SUBTYPE_PRESENTATION_INTERVENTION_NAME;
      case "DataPaper":
      case "JournalArticle":
        return AouConstants.DEPOSIT_SUBTYPE_ARTICLE_SCIENTIFIQUE_NAME;
      case "Dissertation":
        return AouConstants.DEPOSIT_SUBTYPE_THESE_NAME;
      case "Journal":
        return AouConstants.DEPOSIT_SUBTYPE_NUMERO_DE_REVUE_NAME;
      case "Preprint":
        return AouConstants.DEPOSIT_SUBTYPE_PREPRINT_NAME;
      case "Report":
        return AouConstants.DEPOSIT_SUBTYPE_RAPPORT_DE_RECHERCHE_NAME;
      default:
        return null;
    }
  }
}
